/*
  CREATURE.C
  ERKKI-enginen oliorutiinit: polioluokat & niiden ja aliolioiden ksittely.

  Versio 0.8.5

  Tarmo Korpela 2005
*/

#include "erkki.h"    /* Rakenteet ja yleismaailmalliset funktiot */
#include "creature.h" /* Esineistn funktioiden esittely & vakiot */

#ifdef __cplusplus /* C++ -ympristn C-castaus */
extern "C" {
#endif

/* Enginen osoittimet asennettuihin oliopaletteihin. */
CREATUREPALETTE * ERKKI_CREATURES[MAX_CR_PALETTES];
/*
  ERKKI_CR_COUNT: montako oliotaulukkoa asennettu
  MAX_CR_NUM: asennettujen olioiden lukumr yhteens
  ERKKI_CREATURES[]: tietyn taulukon olioiden lukumr
*/
signed int ERKKI_CRCOUNT = 0, MAX_CR_NUM = 0, ERKKI_CREATURES_LKM[MAX_CR_PALETTES];

/*
  Palauttaa asennettujen olioiden lukumrn
*/
signed int ERKKI_MaxCreature(){
  return MAX_CR_NUM;
}

/*
  Etsii kentlt CREATURESTACK:n kahvan, joka osoittaa olioon KUKA.
*/
signed int CreatureHandle(KENTTA * kentta, unsigned int kuka){
unsigned int k;
CREATURESTACK * tmp;
  for(k=0;k<kentta->MaxCreature;k++){
    tmp = ERKKI_creature(kentta, k);
    if(tmp == NULL)return ERKKI_ERROR;
    if(tmp->oliotyyppi == kuka && (tmp->liput & Cr_D_On))return k;
  }
  return ERKKI_NOFOUND; /* ei lydy ko. oliota */
}

/*
  Hakee ensimmisen vapaan oliokahvan kentlt, trai dynaamisessa moodissa
  viimeisimmn varatun olion ID:n (-1 = ei olioita)
*/
signed int FindFreeCreatureHandle(KENTTA * kentta){
int x = 0;
CREATURESTACK * tmp = kentta->Oliot[0];
  if(ERKKI_FLAGS() & ERKKI_DYNAMIC){ /* dynaaminen  - palauttaa viimeisimmn varatun */
    while(tmp){tmp = tmp->next; x++;}
    return x - 1; /* -1, jos ei lainkaan varattuja */
  }else{ /* Taulukoitu */
    for(x=0;x<kentta->MaxCreature;x++)
      if((kentta->Oliot[x]->liput & Cr_D_On) == 0)return x;
  }
  return ERKKI_NOFOUND; /* ei vapaata */
}

/*
  Katsoo, mihin luokkaan linkki (pjakaja) viittaa.
  Koska kelaa nopeassa if:ss, niin kaikki huomiodaan.
*/
signed int WhatCreaturePalette(unsigned int linkki){
signed int a, b=0;
  for(a=0;a<ERKKI_CRCOUNT;a++){
    b += ERKKI_CREATURES_LKM[a];
    if(linkki < b)return a;
  }
  return ERKKI_NOFOUND;
}

/*
  Purkaa pjakajan: monesko omassa listassaan? Eli jos esim. 1.:ss paletissa on 6 oliota, toisessa 4
  ja 3.:ssa 7, niin 8. olio on listassa 2 (WhatCreaturePalette), ja linkiss 1 (asteikolla [0,..])
*/
signed int WhoInCreaturePalette(unsigned int kuka){
signed int a,b=WhatCreaturePalette(kuka),s=0;
  for(a=0;a<b;a++)s += ERKKI_CREATURES_LKM[a];
  /* nyt s:ss on kaikkien edeltvien listojen lkm:t. */
  return kuka - s;
}

signed int ERKKI_CountCreatures(KENTTA * kentta){
signed int c = 0, x;
CREATURESTACK * tmp = kentta->Oliot[0];
  if(ERKKI_FLAGS() & ERKKI_DYNAMIC){
    while(tmp != NULL){c++; tmp = tmp->next;}
  }else{
    for(x=0;x<kentta->MaxCreature;x++)
      if(kentta->Oliot[x]->liput & Cr_D_On)c++;
  }
  return c;
}

inline int ERKKI_getcreature(KENTTA *kentta, int x, int y){
  return kentta->Kentta[x+(y*kentta->w)].olio;
}

signed int GetCreatureByName(char * nimi){
signed int a, b;
  for(a=0; a<MAX_CR_NUM; a++){
    if(strcmp(CreatureNameByType(a), nimi)==0)return a;
  }
  return -1; //Muuten...
}

/*
  Kopioi olion sislln toiseen olioon. KENTTA-muuttujasta katsotaan vain datan mr.
  Sijaintia ei kopioida! Kytetn vain yleisen datan siirtoon!
*/
int CopyCreature(KENTTA * kentta, CREATURESTACK * minne, CREATURESTACK * mista){
int a, b;
  minne->elo = mista->elo; minne->mana = mista->mana; minne->liput = mista->liput;
  minne->oliotyyppi = mista->oliotyyppi;
  /* data liikkuu mukana... */
  for(a=0;a<kentta->MaxCrData;a++)minne->data[a] = mista->data[a];
}

/*
  Asentaa yhden oliolistan engineen. Parametri koko on koko taulukon koko;
  kutsuttaessa staattiselle taulukolle ..._palette(taulu, sizeof(taulu));
  ja dynaamiselle: ...palette(lista, sizeof(CREATUREPALETTE) * alkioiden_maara);
*/
signed int ERKKI_install_creaturepalette(CREATUREPALETTE * lista, size_t koko){
int a;
  if(ERKKI_CRCOUNT>=MAX_CR_PALETTES){
    printf("Varoitus: yritys asentaa liikaa oliopaletteja (MAX=%i)\n", MAX_CR_PALETTES);
    return ERKKI_ERROR;
  }
  ERKKI_CREATURES[ERKKI_CRCOUNT] = lista;
  ERKKI_CREATURES_LKM[ERKKI_CRCOUNT] = (int)(koko/sizeof(CREATUREPALETTE));
  //Nollaus jo tll...
  for(a=0;a<ERKKI_CREATURES_LKM[ERKKI_CRCOUNT];a++)
    ERKKI_CREATURES[ERKKI_CRCOUNT][a].inv = NULL;
  MAX_CR_NUM += ERKKI_CREATURES_LKM[ERKKI_CRCOUNT];
  ERKKI_CRCOUNT ++;
  return ERKKI_OK;
}

/*
  Vapauttaa oliopaletit ja nollaa olioiden lukumrt.
*/
signed int ERKKI_release_creaturepalettes(){
int a;
  for(a=0;a<ERKKI_CRCOUNT;a++){
    ERKKI_CREATURES[ERKKI_CRCOUNT] = NULL;
    ERKKI_CREATURES_LKM[ERKKI_CRCOUNT] = 0;
  }
  ERKKI_CRCOUNT = 0;
}

/*
  Sijoittaa dataa - kuka on MAIN_CREATURE eli polioluokka.
*/
int ERKKI_init_creature_(signed int kuka, signed int * filter, size_t koko, INVENTORYCLASS * var){
int k,a,ali;
  /*
    Varusteet menevt -1:ksi, koska kyseess on initialisointi :)
  */
  k = WhatCreaturePalette(kuka);
  ali = WhoInCreaturePalette(kuka);
  ERKKI_CREATURES[k][ali].koko = (int)(koko/sizeof(int));
  ERKKI_CREATURES[k][ali].filter = (signed int*)malloc(sizeof(int)*ERKKI_CREATURES[k][ali].koko);
  ERKKI_CREATURES[k][ali].varusteet = (signed int*)malloc(sizeof(int)*ERKKI_CREATURES[k][ali].koko);
  for(a=0;a<ERKKI_CREATURES[k][ali].koko;a++){
    ERKKI_CREATURES[k][ali].filter[a] = filter[a];
    ERKKI_CREATURES[k][ali].varusteet[a] = -1;
  }
  ERKKI_CREATURES[k][ali].inv = var;
  return ERKKI_OK;
}

int ERKKI_free_creature(signed int kuka){
signed int k, ali;
  k = WhatCreaturePalette(kuka);
  ali = WhoInCreaturePalette(kuka);
  free(ERKKI_CREATURES[k][ali].filter);
  free(ERKKI_CREATURES[k][ali].varusteet);
  ERKKI_CREATURES[k][ali].koko = 0;
  ERKKI_CREATURES[k][ali].inv = NULL;
  return ERKKI_OK;
}

char * CreatureNameByLoc(KENTTA * kentta, signed int x, signed int y){
  return CreatureNameByHandle(kentta, kentta->Kentta[(x)+(y)*kentta->w].olio);
}

char * CreatureNameByHandle(KENTTA * kentta, signed int kuka){
  if( kuka < 0  || kuka >= kentta->MaxCreature)return NULL;
  return CreatureNameByType(kentta->Oliot[kuka]->oliotyyppi);
}

char * CreatureNameByType(signed int kuka){
int k, ali;
  k = WhatCreaturePalette(kuka);
  ali = WhoInCreaturePalette(kuka);
  if(k>=0)return ERKKI_CREATURES[k][ali].nimi;
  return NULL;
}

/*
  Liikauttaa oliota. Kuka on handleri.
*/
int MoveCreature(KENTTA * kentta, signed int kuka, signed int xm, signed int ym){
signed int x, y, a;
  if(kuka < 0 || kuka >= kentta->MaxCreature)return ERKKI_ERROR;
  if(!xm&&!ym)return ERKKI_OK;
  CreaturePos(kentta, kuka, &x, &y); //otetaan sijainti
  if(CreatureData(kentta, kuka, GET_C_FLAG, 0, SUB_CREATURE) & Cr_D_Stuck)return Este_Kiinni;
  a = ERKKI_onkoeste(kentta, x+xm, y+ym, 0);
  if(a != ERKKI_OK){ //este!
    return a; /* Palautetaan tutkittavaksi */
  }else{
    PosCreature(kentta, kuka, x+xm, y+ym); //ja pivitetn koordinaatit (poistaa samalla ed. kohdasta)
    if(kentta->Kentta[x+xm+(y+ym)*kentta->w].ansa >= 0){ //IIKS
      return ERKKI_TRAP; /* Olio astui ansaan */
    }
  }
  return ERKKI_OK;
}

/* Sijoittaa olion kuka (handler) kohtaan x, y */
int PosCreature(KENTTA * kentta, signed int kuka, signed int x, signed int y){
CREATURESTACK * tmp;
  if(kuka<0 || kuka>=kentta->MaxCreature)return ERKKI_ERROR;
  tmp = ERKKI_creature(kentta, kuka);
  RemoveCreature(kentta, kuka); /* Poistetaan edellisest kohdasta */
  tmp->x = x;
  tmp->y = y;
  kentta->Kentta[(x)+(y)*kentta->w].olio = kuka; /* Tnnekin tieto */
  return ERKKI_OK;
}

/* Nappaa olion kuka (hanska) sijainnin. */
int CreaturePos(KENTTA * kentta, signed int kuka, signed int *x, signed int *y){
CREATURESTACK * tmp;
  if(kuka<0 || kuka>=kentta->MaxCreature)return ERKKI_ERROR;
  tmp = ERKKI_creature(kentta, kuka);
  if(tmp == NULL)return ERKKI_ERROR;
  *x = tmp->x;
  *y = tmp->y;
  return ERKKI_OK;
}

/* Poistaa olion kentlt (handle) */
void RemoveCreature(KENTTA * kentta, signed int kuka){
signed int x, y;
CREATURESTACK * tmp;
  if(kuka<0 || kuka > kentta->MaxCreature)return;
  tmp = ERKKI_creature(kentta, kuka);
  if(tmp == NULL)return;
  /* Nykyinen sijainti */ 
  x = tmp->x; y = tmp->y;
  /* Pyyhitn */
  if(NOTINMAP(kentta, x, y)); else kentta->Kentta[x+((y)*kentta->w)].olio = -1;
  tmp->x = -1;
  tmp->y = -1;
}

/* Kuten Remove, mutta poistaa slotin kytst (kuoli) */
void DeleteCreature(KENTTA * kentta, signed int kuka){
signed int x, y;
CREATURESTACK * tmp, * upr, * lwr;
  if(kuka < 0 || kuka >= kentta->MaxCreature)return;
  tmp = ERKKI_creature(kentta, kuka);
  /* Nykyinen sijainti */
  if(tmp == NULL)return;
  x = tmp->x; y = tmp->y;
  if(NOTINMAP(kentta, x, y)); else kentta->Kentta[x+((y)*kentta->w)].olio = -1;
  /* Pyyhitn */
  if(ERKKI_FLAGS() & ERKKI_DYNAMIC){
    /* Yl- ja alapuoliset oliot */
    upr = ERKKI_creature(kentta, kuka+1);
    lwr = ERKKI_creature(kentta, kuka-1);
    /* Kurotaan vli kiinni */
    if(lwr)lwr->next = upr;
    /* Heivataan bittiavaruuteen */
    ERKKI_UnAllocateCreature(tmp);
  }else{
    tmp->x = -1; 
    tmp->y = -1;
    tmp->elo = -1;
    tmp->mana = -1;
    /* Plt pois ellei jo ole */
    tmp->liput = 0;
  }
}

/* 
  Liikuttaa oliota kuka askelen suuntaan, joka on puskurissa. kuka on handleri. 
  Jos suunta = -1, niin liikutaan sattumanvaraiseen suuntaan.
*/
int StepCreature(KENTTA * kentta, signed int kuka){
signed int xm=0, ym=0, a;
  if(kuka < 0 || kuka >= kentta->MaxCreature)return ERKKI_ERROR;
  for(a=0;a<Max_Motion_Step;a++)
    if(kentta->Oliot[kuka]->Motion[a] < 0)break;
  if(kentta->Oliot[kuka]->Motion[0] < 0){
  }else{ //kohdassa a-1 on liikkeen suunta.
    xym_by_dir(kentta->Oliot[kuka]->Motion[a-1], &xm, &ym);
    kentta->Oliot[kuka]->Motion[a-1] = -1; //Nollaus
    kentta->Oliot[kuka]->Motion[Max_Motion_Step-1] = -1;
  }
  return MoveCreature(kentta, kuka, xm, ym);
}

/* Kuin yll, mutta liikauttaa kaikkia olioita */
int StepCreatures(KENTTA * kentta){
int a;
  for(a=0;a<kentta->MaxCreature;a++){
    if( StepCreature(kentta, a) == ERKKI_ERROR)return ERKKI_ERROR;
  }
  return ERKKI_OK;
}

/*
*/
signed int ERKKI_PutCreatureToMap(KENTTA * kentta, CREATURESTACK * olio, signed int x, signed int y){
CREATURESTACK * tmp;
signed int handle;
  if(olio == NULL)return ERKKI_ERROR;
  /* Vapaa alkio, tai viimeinen varattu */
  handle = FindFreeCreatureHandle(kentta);
  if(NOTINMAP(kentta, x, y))return ERKKI_ERROR;
  /* Onko kohdassa jo olio? */
  if(ERKKI_getcreature(kentta, x, y) >= 0)return ERKKI_NOROOM;
  /* Jos ei tilaa... */
  if(!(ERKKI_FLAGS() & ERKKI_DYNAMIC) && handle == ERKKI_NOFOUND)return ERKKI_ERROR;
  tmp = ERKKI_creature(kentta, handle); /* Suora osoitin vapaaseen/viimeiseen olioon */
  if(ERKKI_FLAGS() & ERKKI_DYNAMIC){
    /* NULL, jos olioita ei ole jonossa. */
    if(tmp == NULL){
      kentta->Oliot[0] = ERKKI_AllocateCreature(kentta); tmp = kentta->Oliot[0];
      if(tmp == NULL)return ERKKI_ERROR;
    }else{
      tmp->next = ERKKI_AllocateCreature(kentta);
      if(tmp->next == NULL)return ERKKI_ERROR;
      /* Hyptn juuri varattuun palaan... */
      tmp = tmp->next;
    }
    /* Ei seuraavaa oliota */
    tmp->next = NULL;
    handle++;
  }
  /* Kopioidaan sislt */
  CopyCreature(kentta, tmp, olio);
  if(PosCreature(kentta, handle, x, y) != ERKKI_OK)return ERKKI_ERROR; /* Laitetaan kartalle */
  return handle;
}

/*
  Asettaa olion kartalle, tysiss voimissa. hahmo on polio.
  Palauttaa kahvan, johon hahmo laitettiin (0,1,2...). Muuten ERKKI_ERROR.
*/
int PutCreatureToMap(KENTTA * kentta, signed int hahmo, signed int x, signed int y){
signed int handle;
CREATURESTACK * olio = ERKKI_AllocateCreature(kentta);
  if(olio == NULL)return ERKKI_ERROR;
  olio->oliotyyppi = hahmo;
  olio->elo = CreatureData(kentta, hahmo, GET_C_ELOMAX, 0, MAIN_CREATURE);
  olio->mana = CreatureData(kentta, hahmo, GET_C_MANAMAX, 0, MAIN_CREATURE);
  olio->liput = Cr_D_On;
  handle = ERKKI_PutCreatureToMap(kentta, olio, x, y);
  ERKKI_UnAllocateCreature(olio);
  return handle;
}

/*
  Modify - asettaa datan, jos mahdollista.
*/
int ModifyCreature(KENTTA * kentta, signed int kuka, int data, int data2, signed int sdata, int jako){
signed int k, r, a;
  if(jako == SUB_CREATURE){
    if(kuka < 0 || kuka >= kentta->MaxCreature)return 0;
    k = WhatCreaturePalette(kentta->Oliot[kuka]->oliotyyppi);
  }else{ /* Tutkitaan polioluokkia kuka */
    if(kuka < 0)return 0;
    k = WhatCreaturePalette(kuka); //0...
    a = WhoInCreaturePalette(kuka);
  }
  if(k<0)return 0;

  if(jako == SUB_CREATURE){ /* Hyvin vhn muokattavissa */
    switch(data){
      case GET_C_ELO: kentta->Oliot[kuka]->elo = sdata; break;
      case GET_C_MANA:kentta->Oliot[kuka]->mana = sdata; break;
      case GET_C_FLAG: kentta->Oliot[kuka]->liput = (unsigned int)sdata; break;
      case GET_C_DATA: 
        if(data2 >= 0 && data2 < kentta->MaxCrData)kentta->Oliot[kuka]->data[data2] = sdata;
        break;
      default: break;
    };  
  }else{ //Poliot
    switch(data){
      case GET_C_ELOMAX: ERKKI_CREATURES[k][a].elomax = sdata; break;
      case GET_C_MANAMAX: ERKKI_CREATURES[k][a].manamax = sdata; break;
      case GET_C_SLOT:
        for(r=0;r<ERKKI_CREATURES[k][a].koko;r++){
          if(ERKKI_CREATURES[k][a].filter[r] == data2){ERKKI_CREATURES[k][a].varusteet[r] = sdata; return ERKKI_OK;}
        }
        return ERKKI_NOFOUND;
      case GET_C_SLOT_I:
        if(data2 < 0 || data2 >= ERKKI_CREATURES[k][a].koko)return ERKKI_NOFOUND;
        ERKKI_CREATURES[k][a].varusteet[data2] = sdata; 
        break;
      case GET_C_VOIMA: ERKKI_CREATURES[k][a].voima = sdata; break;      
      case GET_C_RANGE: ERKKI_CREATURES[k][a].range = sdata; break;      
      case GET_C_SUOJAUS: ERKKI_CREATURES[k][a].suojaus = sdata; break;
      case GET_C_FLAG: ERKKI_CREATURES[k][a].liput = (unsigned int)sdata; break;      
      case GET_C_DATA:
        if(data2 >= 0 && data2 < MaxOlioDataNum)ERKKI_CREATURES[k][a].data[data2] = sdata;
        break;
      default: break;
    };  
  }
  return ERKKI_OK;
}

/* 
  kuka on handleri. Esine on mys handleri (varastoon/kentlle)
  Laittaa tietylle oliolle kamaa plle. Jos st = ... 
  EQ_POIS, vanha tiputetaan maastoon. 
  EQP_KORVAA, vanha pyyhitn veke.
  Kuten muistamme, oliot eivt voi "kantaa" varustetta, jos inv == NULL.

  KESKEN!!

*/
signed int EquipCreature(KENTTA * kentta, signed int kuka, signed int esine, char st, int block){
  return ERKKI_OK;
}

/*
  Vahingoittaa oliota. Poistaa kentlt, jos kuolee. kuka on handleri.
*/
signed int DamageCreature(KENTTA * kentta, signed int kuka, signed int vahinko){
signed int elo, emax;
  if(kuka < 0 || kuka > kentta->MaxCreature)return ERKKI_ERROR;
  elo = CreatureData(kentta, kuka, GET_C_ELO, NON_INFO, SUB_CREATURE);
  emax = CreatureData(kentta, kentta->Oliot[kuka]->oliotyyppi, GET_C_ELOMAX, NON_INFO, MAIN_CREATURE);
  elo -= vahinko;
  if(elo > emax && vahinko < 0)elo = emax; /* Parannettaessa */
  ModifyCreature(kentta, kuka, GET_C_ELO, NON_INFO, elo, SUB_CREATURE);
  if(elo <= 0){
    DeleteCreature(kentta, kuka); /* Jees poks moi moi... */
    return ERKKI_KILLED;
  }
  return ERKKI_OK;
}

/*
  Poistaa oliolta manaa. Jos < 0, niin = 0
*/
signed int ExHaustCreature(KENTTA * kentta, signed int kuka, signed int burn){
signed int mana;
  if(kuka < 0 || kuka > kentta->MaxCreature)return ERKKI_ERROR;
  mana = CreatureData(kentta, kuka, GET_C_MANA, NON_INFO, SUB_CREATURE);
  mana -= burn;
  if(mana < 0)mana=0;
  ModifyCreature(kentta, kuka, GET_C_MANA, NON_INFO, mana, SUB_CREATURE);
  return ERKKI_OK;
}

/*
  Liitt liikkeen oliolle (suuntana)
*/
int AddStepCreature(KENTTA *kentta, signed int kuka, char suunta){
signed int a;
  if(kuka < 0 || kuka >= kentta->MaxCreature)return ERKKI_ERROR;
  for(a=0;a<Max_Motion_Step;a++)
    if(kentta->Oliot[kuka]->Motion[a] < 0)break;
  if(a<Max_Motion_Step){ //a == MaxMS, jos ei tilaa
    kentta->Oliot[kuka]->Motion[a] = suunta;
  }else return ERKKI_NOROOM;
  return a;
}

/*
  Tehokkaampi tapa liikkeen laittamiseksi - suora jono.
*/
int SetStepCreature(KENTTA *kentta, signed int kuka, char * suunta){
signed int a;
  if(kuka < 0 || kuka >= kentta->MaxCreature)return ERKKI_ERROR;
  if(memcpy(&kentta->Oliot[kuka]->Motion, suunta, Max_Motion_Step)==NULL)return ERKKI_ERROR;
  return ERKKI_OK;
}

/* 
  Palauttaa oliosta halutun datan. Jos jako on SUB, kuka on handle, muuten pluokka. 
*/
int CreatureData(KENTTA * kentta, signed int kuka, int data, int data2, int jako){
signed int k, r, a;
  if(jako == SUB_CREATURE){
    if(kuka < 0 || kuka >= kentta->MaxCreature)return 0;
    k = WhatCreaturePalette(kentta->Oliot[kuka]->oliotyyppi);
  }else{ /* Tutkitaan polioluokkia kuka */
    if(kuka < 0)return 0;
    k = WhatCreaturePalette(kuka); //0...
    a = WhoInCreaturePalette(kuka);
  }
  if(k<0)return 0;
  if(jako == SUB_CREATURE){
    switch(data){
      case GET_C_KUKA: return kentta->Oliot[kuka]->oliotyyppi; 
      case GET_C_ELO: return kentta->Oliot[kuka]->elo; 
      case GET_C_MANA: return kentta->Oliot[kuka]->mana; 
      case GET_C_FLAG: return kentta->Oliot[kuka]->liput;
      case GET_C_DATA:
        if(data2 >= 0 && data2 < kentta->MaxCrData)return kentta->Oliot[kuka]->data[data2];
    };
  }else{ //Polioluokka
    switch(data){
      case GET_C_ELO: return ERKKI_CREATURES[k][a].elomax;
      case GET_C_MANA: return ERKKI_CREATURES[k][a].manamax;
      case GET_C_ELOMAX: return ERKKI_CREATURES[k][a].elomax;
      case GET_C_MANAMAX: return ERKKI_CREATURES[k][a].manamax;
      case GET_C_SLOT:
        for(r=0;r<ERKKI_CREATURES[k][a].koko;r++){
          if(ERKKI_CREATURES[k][a].filter[r] == data2)return r;
        }
        return ERKKI_NOFOUND;
      case GET_C_SLOT_I:
        if(data2 < 0 || data2 >= ERKKI_CREATURES[k][a].koko)return ERKKI_NOFOUND;
        else return ERKKI_CREATURES[k][a].varusteet[data2];
      case GET_C_VOIMA: return ERKKI_CREATURES[k][a].voima;
      case GET_C_RANGE: return ERKKI_CREATURES[k][a].range;
      case GET_C_SUOJAUS: return ERKKI_CREATURES[k][a].suojaus;
      case GET_C_FLAG: return ERKKI_CREATURES[k][a].liput;
      case GET_C_VAR: return ERKKI_CREATURES[k][a].koko;
      case GET_C_MIND: return ERKKI_CREATURES[k][a].mel1;
      case GET_C_MAXD: return ERKKI_CREATURES[k][a].mel2;
      case GET_C_DATA:
        if(data2 >= 0 && data2 < MaxOlioDataNum)return ERKKI_CREATURES[k][a].data[data2];
        break;
    };  
  }
  return ERKKI_OK;
}

/*
  Tutkii pikaisesti aliolion kuka esinetilaa item. Jos olioluokalle on asennettu varasto,
  niin katsotaan onko ko. tilassa linkki varaston esineeseen; jos taas
  varastoa ei ole niin katsotaan onko linkki >= 0 jolloin asennettu filtteri toimii
  esineluokkana.
  Palautusarvo: esineen pluokka, t:hen sijoitetaan linkki
*/
signed int BasicItemLink(KENTTA * kentta, signed int kuka, signed int item, signed int * t){
signed int a, luokka, mones;
  *t = -1;
  if(kuka < 0 || kuka >= kentta->MaxCreature || item < 0)return ERKKI_ERROR;
  luokka = WhatCreaturePalette(kentta->Oliot[kuka]->oliotyyppi);
  mones = WhoInCreaturePalette(kentta->Oliot[kuka]->oliotyyppi);
  if(luokka<0)return ERKKI_ERROR;
  /* Onko olioluokalla inventory? */
  if(ERKKI_CREATURES[luokka][mones].inv == NULL){
    if(item < 0 || item >= ERKKI_CREATURES[luokka][mones].koko)return ERKKI_ERROR;
    *t = ERKKI_CREATURES[luokka][mones].varusteet[item];
    return ERKKI_CREATURES[luokka][mones].filter[item];
  }else{ /* On inventory; linkki kulkee sinne... */
    item = ERKKI_CREATURES[luokka][mones].varusteet[item];
    if(item < 0 || item >= ERKKI_CREATURES[luokka][mones].inv->ItemNum)return ERKKI_ERROR;
    *t = ERKKI_CREATURES[luokka][mones].inv->Items[item].Esine.linkki;
    return ERKKI_CREATURES[luokka][mones].inv->Items[item].Esine.mika_esine;
  }
  /* Pstiink tnne? */
  return ERKKI_ERROR;
}

/* 
  Hakee kaikkien olioiden handlerit alueelta. 
  Palauttaa lukumrn.
*/
int ERKKI_GetCreatures(KENTTA * kentta, signed int x1, signed int y1, signed int w, signed int h, 
                       signed int * tulos, int tyyli){
signed int x, y, c=0, k;
  for(y=y1;y<y1+h;y++){
    for(x=x1;x<x1+w;x++){
      if( x >= 0 && y >= 0 && y < kentta->h && x < kentta->w ){
        k = x+(y*kentta->w);
        if(kentta->Kentta[k].olio >= 0){
          if( (tyyli == GET_VISIBLE && (kentta->Kentta[k].data & ERKKI_LOS_LIGHT) > 0) ||
              (tyyli == GET_ALL)
             ){
            tulos[c] = kentta->Kentta[k].olio;
            c++;
          }
        }
      }
    }
  }
  return c;
}

/* 
  Hakee kaikkien olioiden handlerit alueelta. Palauttaa lukumrn.
  Kuten ERKKI_GetCreatures, mutta hakee alueen kehmisesti ->
*/
int ERKKI_GetCreatures_ring(KENTTA * kentta, signed int x1, signed int y1, signed int r, signed int * tulos, int tyyli){
signed int a, z, c=0, k, Koko = kentta->w * kentta->h;
signed int xx,yy;
  //keskikohta
  if(kentta->Kentta[x1 + y1 * kentta->w].olio >= 0){
    tulos[c++] = kentta->Kentta[x1 + y1 * kentta->w].olio;
  }
  //muut alueet
  for(k=1;k<r;k++){
    //Ensin yl- ja alaosa:
    for(a=0; a < (2*k)+1; a++){
      xx = x1 - k + a; yy = y1 - k;
      z = xx + yy * kentta->w; //ylosan sijainti
      if(!(NOTINMAP(kentta, xx, yy)))
        if(kentta->Kentta[z].olio >= 0)
          if( (tyyli == GET_VISIBLE && (kentta->Kentta[z].data & ERKKI_LOS_LIGHT) > 0) ||
              (tyyli == GET_ALL)
             ) tulos[c++] = kentta->Kentta[z].olio;
      yy += (k * 2);
      z = xx + yy * kentta->w; //ylosan sijainti
      if(!(NOTINMAP(kentta, xx, yy)))
        if(kentta->Kentta[z].olio >= 0)
          if( (tyyli == GET_VISIBLE && (kentta->Kentta[z].data & ERKKI_LOS_LIGHT) > 0) ||
              (tyyli == GET_ALL)
             ) tulos[c++] = kentta->Kentta[z].olio;
    }
    //reunat (oikea & vasen)
    for(a=0; a < (2*k)-1; a++){
      xx = x1 - k; yy = y1 - k + 1 + a;
      z = xx + yy * kentta->w; //vasen reunasijainti
      if(!(NOTINMAP(kentta, xx, yy)))
        if(kentta->Kentta[z].olio >= 0)
          if( (tyyli == GET_VISIBLE && (kentta->Kentta[z].data & ERKKI_LOS_LIGHT) > 0) ||
              (tyyli == GET_ALL)
             ) tulos[c++] = kentta->Kentta[z].olio;
      xx += (k * 2); //puolelta toiselle
      z = xx + yy * kentta->w; //ylosan sijainti
      if(!(NOTINMAP(kentta, xx, yy)))
        if(kentta->Kentta[z].olio >= 0)
          if( (tyyli == GET_VISIBLE && (kentta->Kentta[z].data & ERKKI_LOS_LIGHT) > 0) ||
              (tyyli == GET_ALL)
             ) tulos[c++] = kentta->Kentta[z].olio;
    }
  }
  return c;
}

/* Kirjoittaa olion tiedot tiedostoon. Kuka on pjakaja. */
int ERKKI_writecreature(signed int kuka, FILE * tiedosto){
int k, nlen, a, ali;
  if(tiedosto == NULL)return ERKKI_ERROR;
  k = WhatCreaturePalette(kuka);
  ali = WhoInCreaturePalette(kuka);
  if(k<0||ali<0){
    printf("Virhe: tallennetaan virheellist oliota [%i,%i] tiedostovirtaan.\n\r",k,ali);
    return ERKKI_ERROR;
  }
  nlen = strlen(ERKKI_CREATURES[k][ali].nimi)+1; //+1, pttv merkki mukaan :)
  fwrite(&nlen, sizeof(int), 1, tiedosto);
  fwrite(ERKKI_CREATURES[k][ali].nimi, sizeof(char), nlen, tiedosto);
  /* Elot ym. Koska inv:i ei saa tallentaa tss, niin mennn paloittain. */
  fwrite(&ERKKI_CREATURES[k][ali].elomax, sizeof(int), 1, tiedosto);
  fwrite(&ERKKI_CREATURES[k][ali].manamax, sizeof(int), 1, tiedosto);
  fwrite(&ERKKI_CREATURES[k][ali].mel1, sizeof(int), 1, tiedosto);
  fwrite(&ERKKI_CREATURES[k][ali].mel2, sizeof(int), 1, tiedosto);
  fwrite(&ERKKI_CREATURES[k][ali].suojaus, sizeof(int), 1, tiedosto);
  fwrite(&ERKKI_CREATURES[k][ali].voima, sizeof(int), 1, tiedosto);
  fwrite(&ERKKI_CREATURES[k][ali].range, sizeof(int), 1, tiedosto);
  fwrite(&ERKKI_CREATURES[k][ali].liput, sizeof(int), 1, tiedosto);
  for(a=0;a<MaxOlioDataNum;a++)
    fwrite(&ERKKI_CREATURES[k][ali].data[a], sizeof(int), 1, tiedosto);
  /* Sitten viel varusteet... */
  for(a=0;a<ERKKI_CREATURES[k][ali].koko;a++){
    fwrite(&ERKKI_CREATURES[k][ali].filter[a], sizeof(int), 1, tiedosto);
    fwrite(&ERKKI_CREATURES[k][ali].varusteet[a], sizeof(int), 1, tiedosto);
  }
  if(ERKKI_CREATURES[k][ali].inv != NULL){
    fputc(1, tiedosto);
    for(a=0;a<ERKKI_CREATURES[k][ali].inv->ItemNum;a++)
      fwrite(&ERKKI_CREATURES[k][ali].inv->Items[a], sizeof(ITEMCLASS), 1, tiedosto);
  }else fputc(0, tiedosto);
  return ERKKI_OK;
}

signed int ERKKI_write_single_creature(KENTTA * kentta, CREATURESTACK * olio, FILE * tiedosto){
int t;
  if(tiedosto==NULL)return ERKKI_ERROR;
  /* Koska oliolinkiss on malloc-muttujia, niin pit menn yksitellen. */
  /* 8 intti: elo, mana, oliotyyppi, x,y,kx,ky,liput */
  fwrite(&olio->elo, sizeof(int), 1, tiedosto);
  fwrite(&olio->mana, sizeof(int), 1, tiedosto);
  fwrite(&olio->x, sizeof(int), 1, tiedosto);
  fwrite(&olio->y, sizeof(int), 1, tiedosto);
  fwrite(&olio->kx, sizeof(int), 1, tiedosto);
  fwrite(&olio->ky, sizeof(int), 1, tiedosto);
  fwrite(&olio->liput, sizeof(int), 1, tiedosto);
  fwrite(&olio->oliotyyppi, sizeof(int), 1, tiedosto);
  /* sitten data */
  for(t=0;t<kentta->MaxCrData;t++)fwrite(&olio->data[t], sizeof(int), 1, tiedosto);
  for(t=0;t<kentta->CrItems;t++)fwrite(&olio->varusteet[t], sizeof(int), 1, tiedosto);
  fwrite(&olio->Motion, sizeof(char), Max_Motion_Step, tiedosto);
  return ERKKI_OK;
}

int ERKKI_readcreature(signed int kuka, FILE * tiedosto){
int k, nlen, a, ali;
  if(tiedosto == NULL)return ERKKI_ERROR;
  k = WhatCreaturePalette(kuka);
  ali = WhoInCreaturePalette(kuka);
  if(k<0||ali<0){
    printf("Virhe: ladataan virheellist oliota [%i,%i] tiedostovirrasta.\n\r",k,ali);
    return ERKKI_ERROR;
  }

  fread(&nlen, sizeof(int), 1, tiedosto); //nimen koko
  fread(ERKKI_CREATURES[k][ali].nimi, sizeof(char), nlen, tiedosto);
  /* Elot ym. Koska inv:i ei saa tallentaa tss, niin mennn paloittain. */
  fread(&ERKKI_CREATURES[k][ali].elomax, sizeof(int), 1, tiedosto);
  fread(&ERKKI_CREATURES[k][ali].manamax, sizeof(int), 1, tiedosto);
  fread(&ERKKI_CREATURES[k][ali].mel1, sizeof(int), 1, tiedosto);
  fread(&ERKKI_CREATURES[k][ali].mel2, sizeof(int), 1, tiedosto);
  fread(&ERKKI_CREATURES[k][ali].suojaus, sizeof(int), 1, tiedosto);
  fread(&ERKKI_CREATURES[k][ali].voima, sizeof(int), 1, tiedosto);
  fread(&ERKKI_CREATURES[k][ali].range, sizeof(int), 1, tiedosto);
  fread(&ERKKI_CREATURES[k][ali].liput, sizeof(int), 1, tiedosto);
  for(a=0;a<MaxOlioDataNum;a++)
    fread(&ERKKI_CREATURES[k][ali].data[a], sizeof(int), 1, tiedosto);
  /* Sitten viel varusteet... */
  for(a=0;a<ERKKI_CREATURES[k][ali].koko;a++){
    fread(&ERKKI_CREATURES[k][ali].filter[a], sizeof(int), 1, tiedosto);
    fread(&ERKKI_CREATURES[k][ali].varusteet[a], sizeof(int), 1, tiedosto);
  }
  a = fgetc(tiedosto);
  if(a == 1){
    for(a=0;a<ERKKI_CREATURES[k][ali].inv->ItemNum;a++)
      fread(&ERKKI_CREATURES[k][ali].inv->Items[a], sizeof(ITEMCLASS), 1, tiedosto);
  }else ERKKI_CREATURES[k][ali].inv = NULL;
  return ERKKI_OK;
}

signed int ERKKI_read_single_creature(KENTTA * kentta, CREATURESTACK * olio, FILE * tiedosto){
int t;
  if(tiedosto==NULL)return ERKKI_ERROR;
  /* Koska oliolinkiss on malloc-muttujia, niin pit menn yksitellen. */
  /* 8 intti: elo, mana, oliotyyppi, x,y,kx,ky,liput */
  fread(&olio->elo, sizeof(int), 1, tiedosto);
  fread(&olio->mana, sizeof(int), 1, tiedosto);
  fread(&olio->x, sizeof(int), 1, tiedosto);
  fread(&olio->y, sizeof(int), 1, tiedosto);
  fread(&olio->kx, sizeof(int), 1, tiedosto);
  fread(&olio->ky, sizeof(int), 1, tiedosto);
  fread(&olio->liput, sizeof(int), 1, tiedosto);
  fread(&olio->oliotyyppi, sizeof(int), 1, tiedosto);
  /* sitten data */
  for(t=0;t<kentta->MaxCrData;t++)fread(&olio->data[t], sizeof(int), 1, tiedosto);
  for(t=0;t<kentta->CrItems;t++)fread(&olio->varusteet[t], sizeof(int), 1, tiedosto);
  fread(&olio->Motion, sizeof(char), Max_Motion_Step, tiedosto);
  return ERKKI_OK;
}

#ifdef __cplusplus /* C++ -ympristn C-castaus (loppuu) */
}
#endif