/*
  INVEN.C
  ERKKI-enginen esinevarastorutiinit (laukut yms)
  The inventory routines of ERKKI-engine.

  Versio 0.8.5

  Tarmo Korpela 2005
*/

#include "erkki.h"  /* Rakenteet ja yleismaailmalliset funktiot */
#include "item.h"   /* Esineistn funktioiden esittely & vakiot */
#include "inven.h"  /* Esinevarastojen funktioiden esittely */

#ifdef __cplusplus /* C++ -ympristn C-castaus */
extern "C" {
#endif

/*
  Varataan varastolle muistia - Allocating memory to inventory
  koko = max. mr esineit, jotka varastoon voi tallettaa (taulukko,oodi)
         Max. number of items to be stored
*/
INVENTORYCLASS * ERKKI_INV(unsigned int koko){
INVENTORYCLASS * temp;
  temp = (INVENTORYCLASS*)malloc(sizeof(INVENTORYCLASS*));
  if(temp == NULL)return NULL;
  temp->Items = (ITEMCLASS*)malloc(sizeof(ITEMCLASS)*koko);
  if(temp->Items == NULL){
    free(temp);
    return NULL;
  }
  temp->ItemNum = koko;
  NullItemList(temp);
  return temp;
}

/*
  Vapauttaa varastolle varatun muistin.
*/
void ERKKI_INV_free(INVENTORYCLASS * v){
ITEMCLASS * tmp;
  if(v == NULL)return;
  free(v->Items);
  free(v);
}

/*
  Laskee varastossa olevien esineiden lukumrn.
*/
signed int CountItemsInv(INVENTORYCLASS * varasto){
unsigned int a = 0, b;
  for(b = 0; b < varasto->ItemNum; b++)
    if(varasto->Items[a].Esine.mika_esine >= 0)a++;
  return a;
}

/*
  Hakee ensimmisen vapaan esinetilan varastosta. Palauttaa tilan indeksin tai ERKKI_NOFOUND
  mikli ei lydy.
*/
signed int FreeInvSlot(INVENTORYCLASS * varasto){
int a = 0, r=ERKKI_NOFOUND;
  for(a=0;a<varasto->ItemNum;a++)
    if(varasto->Items[a].Esine.mika_esine == -1){r = a; break;}
  return r;
}

/*
  Tyhjent varaston eli asettaa kaikkien esineiden luokaksi -1.
  (Ja varmuuden vuoksi linkiksi 0).
*/
signed int NullItemList(INVENTORYCLASS *varasto){
signed int a;
  if(varasto == NULL)return ERKKI_ERROR;
  for(a=0;a<varasto->ItemNum;a++){
    varasto->Items[a].Esine.mika_esine = -1;
    varasto->Items[a].Esine.linkki = 0;
    varasto->Items[a].Esine.kunto = 0;
    varasto->Items[a].Esine.data1 = 0;
    varasto->Items[a].Esine.data2 = 0;
  }
  return ERKKI_OK;
}

/*
  Poimii kentlt sijainnista (x,y) esineen nmr. heap alhaalta lukien
  eli alin=0, seuraava 1 jne. Palauttaa
  ERKKI_ERROR mikli tapahtuu virhe,
  ERKKI_NOITEM mikli kohdassa ei ole esineit
  
*/
int PickUpItem(KENTTA * kentta, int x, int y, int heap, INVENTORYCLASS *varasto){
signed int v, a, i, k = x + y*kentta->w, z=0;
  if(NOTINMAP(kentta, x, y))return ERKKI_ERROR; /* Ei olla kartalla :P */
  if(kentta->Kentta[k].esine < 0)return ERKKI_NOITEM;
  i = ItemLink(kentta, x, y, heap); /* Itemin viite kohdassa x,y */
  /* Samanlaiset rahat isketn yhteen */
  if(kentta->Esineet[i].Esine.mika_esine == Esine_Raha){
    z = 1;
    /* Onko toista samanlaista (kulta, hopea, pronssi) rahakasaa? */
    a = ScanInventory(varasto, Esine_Raha, kentta->Esineet[i].Esine.linkki, 0);
    if(a<0){a = FreeInvSlot(varasto);z=0;}
  }else a = FreeInvSlot(varasto);
  if(varasto != NULL){
    if(a < 0)return ERKKI_NOROOM;
    else{
      if(CutItem(kentta, i) != ERKKI_OK)return ERKKI_ERROR; /* Jos ei voida leikata */
      /* Kopioidaan */
      if(z){ //z=1, eli oli samanlainen rahakasa. Listn kuntoa:
        kentta->Esineet[i].Esine.kunto += varasto->Items[a].Esine.kunto;
      }
      memcpy(&varasto->Items[a].Esine, &kentta->Esineet[i].Esine, sizeof(ITEMCLASS));
      DestroyItem(kentta, i); /* Ja tuhotaan lopullisesti kentlt */ 
      return a; /* Palautetaan tila, johon tungettiin */
    }
  }else{
    return ERKKI_ERROR;
  }
  return ERKKI_OK; //Tnne ei tulla koskaan ...
}

/*
  Pudottaa varastosta esineen nmr. link kentlle sijaintiin (x,y)
*/
int DropDownItem(KENTTA * kentta, INVENTORYCLASS * varasto, signed int x, signed int y, signed int link){
int v;
  if(NOTINMAP(kentta, x, y))return ERKKI_ERROR; /* Ei olla kartalla */
  if(varasto == NULL){ //eips
    return ERKKI_ERROR;
  }else{ //varasto on olemassa
    if(link < 0){
      printf("Virhe varastosta kopioinnissa: epkyp viittaus (%i).\n\r", link);
      return ERKKI_ERROR;
    }
    if((v = CopyItemToMap(kentta, &varasto->Items[link].Esine, x, y)) < 0){
      if(v==ERKKI_ERROR)printf("Virhe varastosta kopioinnissa: virhe kentlle kopioinnissa.\n\r");
      getch();
      return v;
    }
    varasto->Items[link].Esine.mika_esine = -1; //nollataan
    return v;
  }
  return ERKKI_ERROR;
}

/*
  Jrjestelee inventoorin aukottomaksi eli asettaa kaikki varatut tilat
  etuphn.
*/
void UnSpaceInventory(INVENTORYCLASS *varasto){
signed int a, b, c;
ITEMSCLASS esine;
  for(a=1;a<varasto->ItemNum;a++){
    if(varasto->Items[a-1].Esine.mika_esine < 0 && varasto->Items[a].Esine.mika_esine >= 0){ //aich... vedetn alaspin
      memcpy(&esine, &varasto->Items[a].Esine, sizeof(ITEMSCLASS));
      c = 0;
      for(b=a-1;b>=0;b--)if(varasto->Items[b-1].Esine.mika_esine >= 0 || b == 0){c=1;break;}
      if(c){
        varasto->Items[a].Esine.mika_esine = -1;
        memcpy(&varasto->Items[b].Esine, &esine, sizeof(ITEMSCLASS));        
      }
    }
  }
}

/*
  Poistaa aukot inventorista, huomioiden varustamisen olioluokkaan.
  Kantaja on pjakaja (0...MaxCreature)
*/
void UnSpaceEquipInventory(KENTTA * kentta, INVENTORYCLASS *varasto, signed int kantaja){
signed int a, b, c;
ITEMSCLASS esine; /* Kytetn esineen datan varastointiin */
  for(a=1;a<varasto->ItemNum;a++){
    if(varasto->Items[a-1].Esine.mika_esine < 0 && varasto->Items[a].Esine.mika_esine >= 0){ //aukko alapuolella
      //esine muistiin...
      memcpy(&esine, &varasto->Items[a].Esine, sizeof(ITEMSCLASS));
      c = 0;
      for(b=a-1;b>=0;b--)if(varasto->Items[b-1].Esine.mika_esine >= 0 || b == 0){c=1;break;}
      if(c){ //lytys parempi tila, b:ss
        if(ERKKI_isequip(kentta, kantaja, a)>0){ //esine on varustettuna, joten pit muuttaa sitkin
          //muunna slotti, jossa esine oli, a:sta b:ksi
          ModifyCreature(kentta, kantaja, GET_C_SLOT_I, ERKKI_isequip(kentta, kantaja, a)-1, b, MAIN_CREATURE);
        }
        varasto->Items[a].Esine.mika_esine = -1; //vapaaksi
        memcpy(&varasto->Items[b].Esine, &esine, sizeof(ITEMSCLASS));
      }
    }
  }
}

/* Kopioi esineen varastoon. Paljon helpompi ja nopeampi tapa tutkia/muokata yks. esinett. */
signed int ERKKI_copy_in_inventory(INVENTORYCLASS * varasto, signed int mika, ITEMSCLASS * esine){
  if(mika < 0 || mika >= varasto->ItemNum)return ERKKI_ERROR;
  memcpy(&varasto->Items[mika].Esine, esine, sizeof(ITEMSCLASS));
  return ERKKI_OK;
}

/* Kopioi esineen varastosta. Paljon helpompi ja nopeampi tapa tutkia/muokata yks. esinett. */
signed int ERKKI_copy_from_inventory(INVENTORYCLASS * varasto, signed int mika, ITEMSCLASS * esine){
  if(mika < 0 || mika >= varasto->ItemNum)return ERKKI_ERROR;
  memcpy(esine, &varasto->Items[mika].Esine, sizeof(ITEMSCLASS));
  return ERKKI_OK;
}

/* 
  Kopioi esineen numero item varastosta v1 v2:een. Jos ei ole tilaa, ei siirret. 
  Kaupankyntiin ehdoton! Rahat kertn yh kasoihin.
*/
signed int ERKKI_inv_to_inv(INVENTORYCLASS * v1, INVENTORYCLASS * v2, signed int item){
signed int a,b,c;
  if(item < 0||item >= v1->ItemNum)return ERKKI_ERROR;
  if(v1->Items[item].Esine.mika_esine == Esine_Raha){
    a = ScanInventory(v2, Esine_Raha, v1->Items[item].Esine.linkki, 0);
    if(a<0)a = FreeInvSlot(v2); /* Ei ollut samanlaista rahakasaa */
    else{ v1->Items[item].Esine.kunto += v2->Items[a].Esine.kunto; }
  }else a = FreeInvSlot(v2);
  if(a<0)return ERKKI_NOROOM;
  /* On tilaa, kopioidaan esine kohtaan [a]. */
  memcpy(&v2->Items[a].Esine, &v1->Items[item].Esine, sizeof(ITEMSCLASS));
  v1->Items[item].Esine.mika_esine = -1;
  return ERKKI_OK;
}

/*
  Etsii jotain tietty esinett/esineryhm varastosta. -1 = mik tahansa.
  ss: mist alkaen.
*/
signed int ScanInventory(INVENTORYCLASS *varasto, signed int ml, signed int ll, signed int ss){
signed int a;
  for(a=ss;a<varasto->ItemNum;a++){
    if(varasto->Items[a].Esine.mika_esine == ml || ml < 0)
      if(varasto->Items[a].Esine.linkki == ll || ll < 0)
        return a;
  }
  return -1;
}

/*
  Helpottaa esineen datan hakemista; tarvitsee vain antaa varasto ja esineen kahva
  sijaintiin varastossa (item).
*/
signed int ERKKI_GetItemDataInv(INVENTORYCLASS * inv, int item, int mita, int mc){
  if(item<0||item>=inv->ItemNum)return 0;
  return GetItemData(inv->Items[item].Esine.mika_esine, inv->Items[item].Esine.linkki, mita, mc);
}

/* 
  Onko esine nmr. item varastosta varustettuna oliolla? Otetaan selv. Kuka on pjakaja. 
  Item:n tytyy olla viittaus inventooriin. (Jos ei oliolla ei ole varustusmahd., niin palauttaa ERROR)
  Jos kantaa niin palauttaa slot+1.
*/
signed int ERKKI_isequip(KENTTA * kentta, signed int kuka, signed int item){
int a, b, k=CreatureData(kentta, kuka, GET_C_VAR, 0, MAIN_CREATURE);
  if(k < 1)return ERKKI_ERROR;
  for(a=0;a<k;a++){
    if(CreatureData(kentta, kuka, GET_C_SLOT_I, a, MAIN_CREATURE) == item)return a+1;
  }
  return ERKKI_OK;
}

/*
  Palauttaa varaston esinetilan linkki esineen perusnimen tai NULL mikli
  ko. tilassa ei ole esinett.
*/
char * ItemNameByBlankInv(INVENTORYCLASS * i, signed int linkki){
  if(linkki < 0 || linkki >= i->ItemNum)return NULL;
  return ItemNameByBlank(i->Items[linkki].Esine.mika_esine, i->Items[linkki].Esine.linkki);
}

#ifdef __cplusplus /* C++ -ympristn C-castaus (loppuu) */
}
#endif