/*
  ITEM.C
  ERKKI-enginen esinerutiinit.

  Versio 0.8.5

  Tarmo Korpela 2005
*/

#include "erkki.h" /* Rakenteet ja yleismaailmalliset funktiot */
#include "item.h"  /* Esineistn funktioiden esittely & vakiot */

#ifdef __cplusplus /* C++ -ympristn C-castaus */
extern "C" {
#endif

/*
  Nimet, jotka ItemString palauttaa. Jrjestys ITEM.H:ssa (ellei ERKKI.H:ssa).
  Eli teksti, jota voidaan kytt esineluokkien rajaukseen esim. inventoryss.
*/
static char * ITEM_STRING[Esine_Max] = { //vakiot lytyvt h-tiedostosta
#ifdef TXT_ENGLISH
"Armor", "Weapon", "Shield", "Helm", "Boots", "Cloak", "Belt", "Gloves", "Else",
"Potion", "Arrow", "Book", "Scroll", "Food", "Money", "Spell", "Ring", "Jewelry"
#else
"Panssari", "Ase", "Kilpi", "Kypr", "Kengt", "Viitta", "Vy", "Hansikkaat", 
"Muu", "Juoma", "Ammus", "Kirja", "Kr", "Ruoka", "Raha", "Loitsu", "Sormus", "Koru"
#endif
};

/* Teksti, jonka ItemNameByLoc palauttaa jos kohdassa useampi kuin 1 esine */
static char * HEAP_STRING = "Kasa tavaraa";

/* Eri esineluokkien esineiden mrt. Lasketaan asennuksen yhteydess. */
signed int ITEM_MAARA[Esine_Max] = {-1};
struct RAW_ITEMCLASS * ERKKI_ESINEET[Esine_Max] = {NULL};

/*
  Asennusfunktiot esineille, eli sijoittaa tietyn esineluokan pointterin
  enginen sisisiin rakenteisiin (lueteltu yll), jonka jlkeen funktioita
  voidaan kytt nimien ja muiden tietojen kermiseen kentlt, varastosta
  tai muualta miss esineit voi olla.
  Lyhyempi muoto on makro ERKKI_install_items(int luokka, void * data);
  joka sijoittaa automaattisesti asennettavan rakenteen koon:
*/
int ERKKI_install_items_(int esineluokka, RAW_ITEMCLASS * data, size_t koko){  
  if(esineluokka < 0 || esineluokka >= Esine_Max){
    printf("Varoitus: tunnistamattoman esineluokan (%i) asennusyritys!\n\r", esineluokka);
    return ERKKI_ERROR;
  }else{
    ERKKI_ESINEET[esineluokka] = data;
    ITEM_MAARA[esineluokka] = (int)(koko/sizeof(RAW_ITEMCLASS));
  }
  return ERKKI_OK;
}

/*
  Palauttaa tietyn esineluokan esineiden mrn.
  Esim. Max_Esine(Esine_Ase) palauttaa aseiksi asennetun rakennetaulukon
  alkioiden lukumrn.
*/
int Max_Esine(int esineluokka){
  if(esineluokka < 0 || esineluokka >= Esine_Max){
    printf("Virhe: tuntemattoman esineluokan (%i) rajamritys!\n\r", esineluokka);
    return ERKKI_ERROR;
  }else
    return ITEM_MAARA[esineluokka];
}

/*
  Palauttaa esineluokan a tunnustekstin (Ase, Panssari, Juoma etc...) tai NULL jos
  luokka on vr.
*/
char * Item_String(signed int a){
  if(a>=0 && a < Esine_Max)
    return ITEM_STRING[a];
  else return NULL;
}

/*
  Asettaa esineen i kohtaan x, y. Esinetila (i) tytyy olla valmisteltuna ennen tt
  (eli mika_esine >= 0, muuten esinett ei huomioida muissa funktioissa).
*/
int PutItemToMap(KENTTA * kentta, signed int i, unsigned int x, unsigned int y){
signed int a,b,k=x+y*kentta->w;
ITEMCLASS * tmp = ERKKI_item(kentta, i);
  if(tmp == NULL){
    /* Esineen kahva (i) ei ole varatulla alueella tai sit ei ole luotu */
    printf("Virhe asetettaessa esinett: epkyp kahva (%i).\n\r", i);
    return ERKKI_ERROR;
  }
  if(kentta->Kentta[k].esine < 0){
    /* Yksinkertaisin tapaus - maassa ei ole esineit, joten esine linkitetn suoraan maanpintaan. */
    kentta->Kentta[k].esine = i;
    tmp->alaketju = -1;
    tmp->ylaketju = -1;
  }else{
    /* On jo esineit -> etsitn viimeisin ja linkitetn siihen */
    a = kentta->Kentta[k].esine; /* Ensimmisen esineen linkki */
    /* Edetn linkitetyst esineest toiseen, kunnes saavuteaan ketjun p
     * eli ylaketju == -1 (ei linkityst) */
    while(1){
      if(a == i){/* Linkitysvirhe - sama esine jo linkitetty */
        printf("Virhe esineen linkityksess: silmukkaviittaus esineess %i.\n\r", a);
        return ERKKI_ERROR;
      }
      b = a;      
      a = kentta->Esineet[a].ylaketju;
      if( a < 0 )break;
    }
    /* nyt b on sen esineen nmro, joka on ketjun pss */
    kentta->Esineet[b].ylaketju = i;
    kentta->Esineet[i].ylaketju = -1;
    kentta->Esineet[i].alaketju = b;
  }
  return ERKKI_OK; //OK
}

/*
  Kopioi esineen *esine kentlle. Etsii automaattisesti vapaan tilan, johon
  asettaa esineen linkin. Palauttaa tilan (i) josta esineeseen viitataan kopioinnin jlkeen.
*/
int CopyItemToMap(KENTTA * kentta, ITEMSCLASS * esine, unsigned int x, unsigned int y){
signed int i;
  if( NOTINMAP(kentta, x, y) ){
    printf("Virhe kopioitaessa esinett: kopiointi kentn ulkopuolelle.\n\r");
    return ERKKI_ERROR;
  }
  if(esine->mika_esine < 0){
    printf("Virhe kopioitaessa esinett: NULL-esine.\n\r");
    return ERKKI_ERROR;
  }
  i = FreeItemSlot(kentta); /* Vapaa tila */
  if(i<0)return ERKKI_NOROOM; /* Ei tilaa - palautetaan virhe*/
  /* Kopioidaan esineen data kentn esinetaulukkoon */
  memcpy(&kentta->Esineet[i].Esine, esine, sizeof(ITEMSCLASS));
  /* Linkitetn uusi esine sijaintiin (x,y) */
  PutItemToMap(kentta, i, x, y);
  return i; /* Palautetaan sijainti */
}

/* Kopioi kentn tilan i esineen rakenteeseen *esine. */
int CopyItemFromMap(KENTTA * kentta, ITEMSCLASS * esine, signed int i){
  if(i<0||i>=kentta->MaxItem)return ERKKI_ERROR;
  memcpy(esine, &kentta->Esineet[i].Esine, sizeof(ITEMSCLASS));
  return ERKKI_OK;
}

/* Palauttaa kentn seuraavan vapaan inventoorislotin. */
int FreeItemSlot(KENTTA * kentta){
int a;
  for(a=0;a<kentta->MaxItem;a++)
    if(kentta->Esineet[a].Esine.mika_esine < 0)return a;
  return -1;
}

/*
  Palauttaa esinelinkin kohdassa x,y
*/
int ERKKI_getitem(KENTTA *kentta, int x, int y){
  return kentta->Kentta[x+(y*kentta->w)].esine;
}

/*
  Kopioi osoittimen *esine sislln kentn esinetilaan i. Palauttaa ERKKI_ERROR
  mikli kopioinnissa tapahtui virhe, muuten ERKKI_OK.
*/
int MakeItemAs(KENTTA * kentta, unsigned int i, ITEMSCLASS * esine){ 
  if(i < 0 || i > kentta->MaxItem - 1)return ERKKI_ERROR;
  if( memcpy(&kentta->Esineet[i].Esine, esine, sizeof(ITEMSCLASS)) == NULL)return ERKKI_ERROR;
  return ERKKI_OK;
}

/*
  Palauttaa kohdassa x,y olevasta pinosta num:n esineen. Jos kohdassa ei esinett,
  palauttaa -1.
*/
signed int ItemLink(KENTTA * kentta, int x, int y, signed int num){
int a, b=-1, c=0, k=x+y*kentta->w;
ITEMCLASS * tmp;
  if(NOTINMAP(kentta, x, y))return ERKKI_ERROR; 
  a = kentta->Kentta[k].esine; /* Lhdetn maasta liikkeelle */
  while(1){
    if(a < 0)break;
    b = a;
    tmp = ERKKI_item(kentta, a);
    a = tmp->ylaketju;//kentta->Esineet[a].ylaketju;
    if(c==num)break;
    c++;
  }
  return b;
}

/*
  Leikkaa esineen pois tilasta i (mutta ei tuhoa sit), eli katkaisee ketjutuksen
  ja kokoaa sen uudelleen yl- ja alapuolelta. Funktio on hieman hidas, sill
  mikli esine on maanpinnalla funktio tytyy etsi sen sijainti ennen
  katkaisuhoitoa.
*/
signed int CutItem(KENTTA * kentta, unsigned int i){
int a, b, x, y, k;
ITEMCLASS * tmp, *yla, * ala;
  if(i < 0 || i > kentta->MaxItem - 1)return ERKKI_ERROR;
  tmp = ERKKI_item(kentta, i);
  if(kentta->Esineet[i].alaketju < 0){ /* Pitisi olla maassa/ei lainkaan... haetaan siis x,y */
    a = 0;
    for(y=0;y<kentta->h;y++){
      for(x=0;x<kentta->w;x++){
        if(kentta->Kentta[x + y*kentta->w].esine == i){a=1;break;}
      }
      if(a)break;
    }
    if(!a)return ERKKI_ERROR; /* Esinett ei ole/se on vrin linkitetty. */
    k = x + (y * kentta->w);
    /* Linkitetn ylpuolella oleva esine maanpintaan */
    kentta->Kentta[k].esine = tmp->ylaketju;//kentta->Esineet[i].ylaketju;
    /* Jos esineen pll oli jotain, muunnetaan se alaketju -1:ksi (maanpinnalla) */
    if(tmp->ylaketju >= 0){
      kentta->Esineet[kentta->Esineet[i].ylaketju].alaketju = -1;
    }
  }else{ /* Ketjun keskell tai loppupss */
    /* Muutetaan linkityst */
    kentta->Esineet[kentta->Esineet[i].alaketju].ylaketju = kentta->Esineet[i].ylaketju;
    if(kentta->Esineet[i].ylaketju >= 0)
      kentta->Esineet[kentta->Esineet[i].ylaketju].alaketju = kentta->Esineet[i].alaketju;
  }
  /* Nm viel pois kytst - poistetaan tydellisesti*/
  kentta->Esineet[i].ylaketju = -1;
  kentta->Esineet[i].alaketju = -1;
  return ERKKI_OK;
}

/*
  'Tyhjent' esineen (mika_esine, linkki = -1).
*/
signed int DestroyItem(KENTTA * kentta, unsigned int i){
  kentta->Esineet[i].Esine.mika_esine = -1;
  kentta->Esineet[i].Esine.linkki = -1;
  return ERKKI_OK;
}

/* 
  Tuhoaa kaikki esineet kohdasta x,y. Palauttaa ERKKI_OK, mikli toiminto onnistui
  ja ERKKI_ERROR jos sijainti on kentn ulkopuolella.
*/
signed int DestroyItemsInLoc(KENTTA * kentta, signed int x, signed int y){
signed int a, k;
  if( NOTINMAP(kentta, x, y) )return ERKKI_ERROR;
  k = x + y * kentta->w;
  a = kentta->Kentta[k].esine;
  if(a>=0){
    kentta->Kentta[k].esine = -1;
  }
  while(1){
    if(a<0)break;
    kentta->Esineet[a].Esine.mika_esine = -1; /* pois plt */
    a = kentta->Esineet[a].ylaketju; /* Seuraava esine */
  }
  return ERKKI_OK;
}

/*
  Esineen nimen rutiinit - nimi linkin ja tyypin mukaan.
  Mikli esineen luokka (mika_esine) tai linkki ei ole kohdallaan,
  palautetaan NULL, muuten pointteri esineen nimeen.
*/
char * ItemNameByBlank(signed int num, signed int linkki){
//  printf("%i,%i", num, linkki); getch();
  if(num < 0 || num > Esine_Max)return NULL; /* Virheellinen luokka */
  if(linkki < 0 || linkki >= Max_Esine(num))return NULL; /* Virh. linkki */
  if(ERKKI_ESINEET[num] != NULL){
    return ERKKI_ESINEET[num][linkki].nimi;
  }else return NULL;
}

/*
  Esineen nimen rutiinit - nimi sijainnin mukaan.
  Mikli sijainnissa ei ole mitn tai sijainti on kentn ulkopuolella, palauttaa NULL.
  Jos sijaintiin on asetettu useampia esineit, palauttaa osoittimen HEAP_STRING-
  merkkijonoon (mritetty item.c:n alussa).
  Muuten palauttaa kohdassa olevan esineen yleisnimen.
*/
char * ItemNameByLoc(KENTTA * kentta, signed int x, signed int y){
int k = x + (y*kentta->w);
  if( NOTINMAP(kentta, x, y) )return NULL;
  if( kentta->Kentta[k].esine < 0 )return NULL;
  else{
    if( kentta->Esineet[kentta->Kentta[k].esine].ylaketju < 0 )
      return ItemNameByBlank(kentta->Esineet[kentta->Kentta[k].esine].Esine.mika_esine, 
                             kentta->Esineet[kentta->Kentta[k].esine].Esine.linkki);
    else return HEAP_STRING;
  }
}

/*
  Esineen nimen rutiinit - nimi tilan (osoitteen) mukaan.
  Mikli tila on taulukon ulkopuolella, palauttaa NULL.
  Muuten palauttaa tilassa i olevan esineen yleisnimen.
*/
char * ItemNameByHandle(KENTTA * kentta, signed int i){
  if(i < 0 || i > kentta->MaxItem)return NULL;
  return ItemNameByBlank(kentta->Esineet[i].Esine.mika_esine, kentta->Esineet[i].Esine.linkki);
}

/*
  Etsii luokan L esineen ID:n perusnimen perusteella. Jos ei lydy, palauttaa ERKKI_NOFOUND.
  Virheellinen luokka aiheuttaa virheen ERKKI_ERROR. Funktio tekee eron suurten ja pienten
  kirjainten vlill.
  Esim. a = GetItemByName(Esine_Ase, "Puukko");
*/
signed int GetItemByName(signed int L, char * nimi){
unsigned int a, b=-1;
  if(L<0||L>=Esine_Max)return ERKKI_ERROR;
  for(a=0;a<ITEM_MAARA[L];a++){
    if(strcmp(ItemNameByBlank(L, a), nimi) == 0){b=a;break;}
  }
  if(b<0)return ERKKI_NOFOUND; else return b;
}

/*
  Laskee kentll olevien esineiden lukumrn (esine on 'pll', jos mika_esine
  >= 0).
*/
signed int CountItems(KENTTA * kentta){
unsigned int a = 0, b;
  for(b = 0; b < kentta->MaxItem; b++)
    if(kentta->Esineet[a].Esine.mika_esine >= 0)a++;
  return a;
}

/*
  Palauttaa esineen meleevahingon.
*/
void GetMeleeDamage(signed int luokka, signed int linkki, signed int *max, signed int *min){
  /* Virheellinen luokka/linkki? */
  if(luokka < 0 || luokka > Esine_Max){*min=0;*max=0;return;}
  if(linkki < 0 || linkki >= Max_Esine(luokka)){*min=0;*max=0;return;}
  if(ERKKI_ESINEET[luokka] != NULL){
    *min = ERKKI_ESINEET[luokka][linkki].v1;
    *max = ERKKI_ESINEET[luokka][linkki].v2;
  }
}

/*
  Palauttaa esineluokasta dataa.
  mc on apudata (esim. monesko data jne)
  Esim. GetItemData(Esine_Ase, 2, GET_I_HINTA, 0) palauttaa asetaulukon aseen nmr. 3 (0,1,2,3,4...)
  hinnan. GET_I_DATA taas vaatii, ett mc on vlill 0...MaxItemDataNum.
*/
signed int GetItemData(signed int ml, signed int l, signed int mita, signed int mc){
  if(l > Max_Esine(ml))return 0;
  switch(mita){
    case GET_I_SUOJAUS: return ERKKI_ESINEET[ml][l].suojaus;
    case GET_I_HINTA: return ERKKI_ESINEET[ml][l].hinta;
    case GET_I_DATA: if(mc>=0&&mc<MaxItemDataNum)return ERKKI_ESINEET[ml][l].data[mc];
    case GET_I_MIND: return ERKKI_ESINEET[ml][l].v1;
    case GET_I_MAXD: return ERKKI_ESINEET[ml][l].v2;
    case GET_I_TEHO: return ERKKI_ESINEET[ml][l].tehokkuus;
    case GET_I_PAINO: return ERKKI_ESINEET[ml][l].massa;
    case GET_I_SIS: return ERKKI_ESINEET[ml][l].sisassa;
    case GET_I_LUOKKA: return ERKKI_ESINEET[ml][l].luokka;
    case GET_I_HYOKKAYS: return ERKKI_ESINEET[ml][l].hyokkays;
    case GET_I_PUOLUSTUS: return ERKKI_ESINEET[ml][l].puolustus;
    default: break;
  };
  return 0;  
}

/*
  Edelliseen helpotus: suoraan kentlt tilan avulla (item).
*/
signed int ERKKI_GetItemData(KENTTA * kentta, int item, int mita, int mc){
  if(item<0||item>=kentta->MaxItem)return 0;
  return GetItemData(kentta->Esineet[item].Esine.mika_esine, kentta->Esineet[item].Esine.linkki, mita, mc);
}

/*
  Kuin GetItemdata, mutta pinvastoin: sijoitetaan tietoa esineluokkaan (parametrit samoja kuin
  GetItemdata:ssa, ja viimeisen viel uusi arvo. Mikli muutosta ei voi tehd, ei tee mitn.
  Palauttaa ERKKI_OK.
*/
signed int ModifyItemData(signed int ml, signed int l, signed int mita, signed int mc, int uusi){
  if(l > Max_Esine(ml))return ERKKI_ERROR;
  switch(mita){
    case GET_I_SUOJAUS:   ERKKI_ESINEET[ml][l].suojaus = uusi; break;
    case GET_I_HINTA:     ERKKI_ESINEET[ml][l].hinta = uusi; break;
    case GET_I_DATA:      if(mc>=0&&mc<MaxItemDataNum)ERKKI_ESINEET[ml][l].data[mc] = uusi; break;
    case GET_I_MIND:      ERKKI_ESINEET[ml][l].v1 = uusi; break;
    case GET_I_MAXD:      ERKKI_ESINEET[ml][l].v2 = uusi; break;
    case GET_I_TEHO:      ERKKI_ESINEET[ml][l].tehokkuus = uusi; break;
    case GET_I_PAINO:     ERKKI_ESINEET[ml][l].massa = uusi; break;
    case GET_I_SIS:       ERKKI_ESINEET[ml][l].sisassa = uusi; break;
    case GET_I_LUOKKA:    ERKKI_ESINEET[ml][l].luokka = uusi; break;
    case GET_I_HYOKKAYS:  ERKKI_ESINEET[ml][l].hyokkays = uusi; break;
    case GET_I_PUOLUSTUS: ERKKI_ESINEET[ml][l].puolustus = uusi; break;
    default: break;
  };
  return ERKKI_OK;
}

/* Kirjoittaa yhden esinetiedon tiedostoon. */
signed int ERKKI_write_single_item(KENTTA * kentta, ITEMSCLASS * esine, FILE * tiedosto){
int t;
  if(tiedosto==NULL)return ERKKI_ERROR;
  /* Koska esinelinkiss on malloc-muttujia, niin pit menn yksitellen. */
  fwrite(&esine->mika_esine, sizeof(int), 1, tiedosto);
  fwrite(&esine->linkki, sizeof(int), 1, tiedosto);
  fwrite(&esine->data1, sizeof(int), 1, tiedosto);
  fwrite(&esine->data2, sizeof(int), 1, tiedosto);
  fwrite(&esine->kunto, sizeof(int), 1, tiedosto);
  return ERKKI_OK;  
}

signed int ERKKI_read_single_item(KENTTA * kentta, ITEMSCLASS * esine, FILE * tiedosto){
int t;
  if(tiedosto==NULL)return ERKKI_ERROR;
  /* Koska esinelinkiss on malloc-muttujia, niin pit menn yksitellen. */
  fread(&esine->mika_esine, sizeof(int), 1, tiedosto);
  fread(&esine->linkki, sizeof(int), 1, tiedosto);
  fread(&esine->data1, sizeof(int), 1, tiedosto);
  fread(&esine->data2, sizeof(int), 1, tiedosto);
  fread(&esine->kunto, sizeof(int), 1, tiedosto);
  return ERKKI_OK;  
}

#ifdef __cplusplus /* C++ -ympristn C-castaus (loppuu) */
}
#endif