/*
  LOS.H
  LOS(Line Of Sight)-routines for ERKKI.

  Version 0.8.5

  Tarmo Korpela 2005

  All of these prosedures use the engines field structure, KENTTA,
  as a parameter. If you're not using engines field structure, you
  can make a clone structure including only the data-member of the
  structure and typecasting it as a KENTTA (defined in ERKKI.H).
*/

#ifndef ERKKI_LOS_H_
# define ERKKI_LOS_H_

/* Include the main engine header (structures etc) - Rakenteet ja vakiot */
#include "erkki.h"

/*
  A couple of constants
  Tarvittavat vakiot
*/
#ifdef ERKKI_LOS_LIGHT
#  define ERKKI_LOS_LIGHT 0XFF /* The light power of location (0-255)*/
#  define ERKKI_LOS_SEEN 0X100 /* Is the location exposed after a LOS-routine */
#endif

/* See explanations from lof.c - Funktiot on selitetty los.c:ss */

/* Simple lookup-routines */
int ERKKI_isexposed(KENTTA *, signed int, signed int);
int ERKKI_isvisible(KENTTA *, signed int, signed int);
void ERKKI_unsee_area(KENTTA *);
void ERKKI_spot_light(KENTTA *, signed int, signed int, unsigned char);

/*
  LOS-routines (in order of decreasing speed)
  LOS-rutiinit (alenevan nopeuden mukaan)
*/
void ERKKI_see_area(KENTTA *, signed int, signed int, signed int, unsigned char);
void ERKKI_see_area_corr(KENTTA *, signed int, signed int, signed int, unsigned char);
void ERKKI_see_area_real(KENTTA *, signed int, signed int, signed int, unsigned char);
void ERKKI_see_area_trig(KENTTA *, signed int, signed int, signed int, unsigned char);

/*
  Single LOS-beam methods
  Yksittiset LOS-sdefunktiot
*/
int InLineOfSight(KENTTA *, signed int, signed int, signed int, signed int);
int InLineOfSight_real(KENTTA *, signed int, signed int, signed int, signed int);

#endif /* ERKKI_LOS_H_ */